/*
 * Copyright © 2021 Fachhochschule für die Wirtschaft (FHDW) Hannover
 *
 * This file is part of ipspiel22-tictactoe-core.
 *
 * ipspiel22-tictactoe-core is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * ipspiel22-tictactoe-core is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * ipspiel22-tictactoe-core. If not, see <http://www.gnu.org/licenses/>.
 */
package de.fhdw.gaming.ipspiel22.tictactoe.core.domain.impl;

import java.util.Objects;

import de.fhdw.gaming.core.domain.GameException;
import de.fhdw.gaming.ipspiel22.tictactoe.core.domain.TicTacToeBoard;
import de.fhdw.gaming.ipspiel22.tictactoe.core.domain.TicTacToeDirection;
import de.fhdw.gaming.ipspiel22.tictactoe.core.domain.TicTacToeField;
import de.fhdw.gaming.ipspiel22.tictactoe.core.domain.TicTacToeFieldState;
import de.fhdw.gaming.ipspiel22.tictactoe.core.domain.TicTacToePosition;

/**
 * Implements {@link TicTacToeField}.
 */
final class TicTacToeFieldImpl implements TicTacToeField {

    /**
     * The board this field belongs to.
     */
    private final TicTacToeBoardImpl board;
    /**
     * The position at which this field is placed on the board.
     */
    private final TicTacToePosition position;
    /**
     * The state of this field.
     */
    private TicTacToeFieldState state;

    /**
     * Creates an TicTacToe field.
     *
     * @param board    The board this field belongs to.
     * @param position The position at which this field is placed on the board.
     * @param state    The state of this field.
     */
    TicTacToeFieldImpl(final TicTacToeBoardImpl board, final TicTacToePosition position,
            final TicTacToeFieldState state) {
        this.board = Objects.requireNonNull(board, "board");
        this.position = Objects.requireNonNull(position, "position");
        this.state = state;
    }

    @Override
    public String toString() {
        return String.format("TicTacToeField[position=%s, state=%s]", this.position, this.state);
    }

    /**
     * {@inheritDoc}
     * <p>
     * Does not compare the boards the fields belong to, respectively.
     */
    @Override
    public boolean equals(final Object obj) {
        if (obj instanceof TicTacToeFieldImpl) {
            final TicTacToeFieldImpl other = (TicTacToeFieldImpl) obj;
            return this.position.equals(other.position) && this.state.equals(other.state);
        }
        return false;
    }

    @Override
    public int hashCode() {
        return Objects.hash(this.position, this.state);
    }

    @Override
    public TicTacToeBoard getBoard() {
        return this.board;
    }

    @Override
    public TicTacToePosition getPosition() {
        return this.position;
    }

    @Override
    public TicTacToeFieldState getState() {
        return this.state;
    }

    @Override
    public boolean hasNeighbour(final TicTacToeDirection direction) {
        return this.board.hasFieldAt(direction.step(this.position));
    }

    @Override
    public TicTacToeField getNeighbour(final TicTacToeDirection direction) {
        if (!this.hasNeighbour(direction)) {
            throw new IllegalArgumentException(String.format("No %s neighbour at %s.", direction, this.position));
        }

        return this.board.getFieldAt(direction.step(this.position));
    }

    @Override
    public void changeState(final TicTacToeFieldState newState) throws GameException {
        final TicTacToeFieldState oldState = this.state;
        final boolean wasEmpty = oldState.equals(TicTacToeFieldState.EMPTY);
        final boolean isEmpty = newState.equals(TicTacToeFieldState.EMPTY);

        if (wasEmpty && isEmpty) {
            throw new GameException(String.format("It is not allowed to do nothing on field at %s.", this.position));
        }
        if (isEmpty) {
            throw new GameException(
                    String.format("It is not allowed to remove a mark from field at %s.", this.position));
        }
        if (!wasEmpty) {
            throw new GameException(String.format("It is not allowed to change a mark on field at %s.", this.position));
        }

        this.state = newState;
    }
}
